from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.utils.http import is_safe_url


def arg(request):
    # ruleid: open-redirect
    return redirect(request.POST.get("next"))


def argh(request):
    # ruleid: open-redirect
    return redirect(request.get_host())


def arghh(request):
    # ruleid: open-redirect
    return redirect(request.method)


def argh2(request):
    # ruleid: open-redirect
    url = request.get_host()
    print("something")
    return redirect(url)


def unsafe(request):
    # ruleid: open-redirect
    url = request.headers.get("referrer")
    print("something")
    return redirect(url)


def safe(request):
    # ok
    url = "https://lmnop.qrs"
    return redirect(url)


def fine(request):
    # ok
    return HttpResponseRedirect("https://google.com")


def unsafe2(request):
    # ruleid: open-redirect
    url = request.POST.get("url")
    return HttpResponseRedirect(url)


def legit(request):
    # ok
    return HttpResponseRedirect(request.get_full_path())


def url_validation(request):
    # ok
    next = request.POST.get("next", request.GET.get("next"))
    if (next or not request.is_ajax()) and not is_safe_url(
        url=next, allowed_hosts=request.get_host()
    ):
        next = "/index"
    response = HttpResponseRedirect(next) if next else HttpResponse(status=204)
    return response


def url_validation2(request):
    # ok
    next = request.POST.get("next", request.GET.get("next"))
    ok = is_safe_url(url=next, allowed_hosts=request.get_host())
    if ok:
        response = HttpResponseRedirect(next) if next else HttpResponse(status=204)
    else:
        response = HttpResponseRedirect("index")
    return response



import json
import subprocess
from pathlib import Path
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import pytest


TESTS_PATH = Path(__file__).parent

MASKED_KEYS = [
    "tool.driver.semanticVersion",
    "results.extra.metavars.*.unique_id.md5sum",
]


def mark_masked(obj, path):
    _mark_masked(obj, path.split("."))


def _mark_masked(obj, path_items):
    key = path_items[0]
    if len(path_items) == 1 and key in obj:
        obj[key] = "<masked in tests>"
    else:
        if key == "*":
            next_obj = list(obj.values())
        else:
            next_obj = obj.get(key)
        if next_obj is None:
            next_objs = []
        elif not isinstance(next_obj, list):
            next_objs = [next_obj]
        else:
            next_objs = next_obj
        for o in next_objs:
            _mark_masked(o, path_items[1:])


def _clean_output_json(output_json: str) -> str:
    """Make semgrep's output deterministic and nicer to read."""
    output = json.loads(output_json)
    for path in MASKED_KEYS:
        mark_masked(output, path)

    return json.dumps(output, indent=2, sort_keys=True)


def _run_semgrep(
    config: Optional[Union[str, Path]] = None,
    *,
    target_name: str = "basic",
    options: Optional[List[Union[str, Path]]] = None,
    output_format: str = "json",
    stderr: bool = False,
) -> str:
    """Run the semgrep CLI.

    :param config: what to pass as --config's value
    :param target_name: which directory within ./e2e/targets/ to scan
    :param options: additional CLI flags to add
    :param output_format: which format to use, valid options are normal, json, and sarif
    :param stderr: whether to merge stderr into the returned string
    """
    if options is None:
        options = []

    options.append("--strict")

    if config is not None:
        options.extend(["--config", config])

    if output_format == "json":
        options.append("--json")
    elif output_format == "sarif":
        options.append("--sarif")

    output = subprocess.check_output(
        ["python3", "-m", "semgrep", *options, Path("targets") / target_name],
        encoding="utf-8",
        stderr=subprocess.STDOUT if stderr else None,
    )

    if output_format in {"json", "sarif"} and not stderr:
        output = _clean_output_json(output)

    return output


@pytest.fixture
def run_semgrep_in_tmp(monkeypatch, tmp_path):
    monkeypatch.setenv("PYTHONPATH", str(TESTS_PATH.parent.resolve()))

    (tmp_path / "targets").symlink_to(Path(TESTS_PATH / "e2e" / "targets").resolve())
    (tmp_path / "rules").symlink_to(Path(TESTS_PATH / "e2e" / "rules").resolve())

    monkeypatch.chdir(tmp_path)

    yield _run_semgrep


def pytest_addoption(parser):
    parser.addoption(
        "--qa",
        action="store_true",
        dest="is_qa",
        default=False,
        help="enable comprehensive QA tests",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "qa: mark tests that only need to run during QA")


def pytest_runtest_setup(item):
    if item.get_closest_marker("qa") and not item.config.getoption("--qa"):
        pytest.skip("skipping QA tests, add --qa flag to run them")



import base64
import os
import shutil
import sys
import tarfile
import tempfile
import time
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from semgrep.constants import DEFAULT_CONFIG_FILE
from semgrep.constants import DEFAULT_CONFIG_FOLDER
from semgrep.constants import DEFAULT_SEMGREP_CONFIG_NAME
from semgrep.constants import ID_KEY
from semgrep.constants import RULES_KEY
from semgrep.constants import SEMGREP_USER_AGENT
from semgrep.constants import YML_EXTENSIONS
from semgrep.util import debug_print
from semgrep.util import is_url
from semgrep.util import print_error
from semgrep.util import print_error_exit
from semgrep.util import print_msg

IN_DOCKER = "SEMGREP_IN_DOCKER" in os.environ
IN_GH_ACTION = "GITHUB_WORKSPACE" in os.environ
REPO_HOME_DOCKER = "/home/repo/"

TEMPLATE_YAML_URL = (
    "https://raw.githubusercontent.com/returntocorp/semgrep-rules/develop/template.yaml"
)

RULES_REGISTRY = {
    "r2c": "https://semgrep.live/c/r/all",
}
DEFAULT_REGISTRY_KEY = "r2c"


def manual_config(pattern: str, lang: str) -> Dict[str, Any]:
    # TODO remove when using sgrep -e ... -l ... instead of this hacked config
    return {
        "manual": {
            RULES_KEY: [
                {
                    ID_KEY: "-",
                    "pattern": pattern,
                    "message": pattern,
                    "languages": [lang],
                    "severity": "ERROR",
                }
            ]
        }
    }


def resolve_targets(targets: List[str]) -> List[Path]:
    base_path = get_base_path()
    return [
        Path(target) if Path(target).is_absolute() else base_path.joinpath(target)
        for target in targets
    ]


def adjust_for_docker(in_precommit: bool = False) -> None:
    # change into this folder so that all paths are relative to it
    if IN_DOCKER and not IN_GH_ACTION and not in_precommit:
        if not Path(REPO_HOME_DOCKER).exists():
            print_error_exit(
                f'you are running semgrep in docker, but you forgot to mount the current directory in Docker: missing: -v "${{PWD}}:{REPO_HOME_DOCKER}"'
            )
    if Path(REPO_HOME_DOCKER).exists():
        os.chdir(REPO_HOME_DOCKER)


def get_base_path() -> Path:
    return Path(".")


def indent(msg: str) -> str:
    return "\n".join(["\t" + line for line in msg.splitlines()])


def parse_config_at_path(
    loc: Path, base_path: Optional[Path] = None
) -> Dict[str, Optional[Dict[str, Any]]]:
    config_id = str(loc)
    if base_path:
        config_id = str(loc).replace(str(base_path), "")
    try:
        with loc.open() as f:
            return parse_config_string(config_id, f.read())
    except FileNotFoundError:
        print_error(f"YAML file at {loc} not found")
        return {str(loc): None}


def parse_config_string(
    config_id: str, contents: str
) -> Dict[str, Optional[Dict[str, Any]]]:
    import yaml  # here for faster startup times

    try:
        return {config_id: yaml.safe_load(contents)}
    except yaml.parser.ParserError as se:
        print_error(f"Invalid yaml file {config_id}:\n{indent(str(se))}")
        return {config_id: None}
    except yaml.scanner.ScannerError as se:
        print_error(f"Invalid yaml file {config_id}:\n{indent(str(se))}")
        return {config_id: None}


def parse_config_folder(
    loc: Path, relative: bool = False
) -> Dict[str, Optional[Dict[str, Any]]]:
    configs = {}
    for l in loc.rglob("*"):
        # Allow manually specified paths with ".", but don't auto-expand them
        if not _is_hidden_config(l.relative_to(loc)) and l.suffix in YML_EXTENSIONS:
            configs.update(parse_config_at_path(l, loc if relative else None))
    return configs


def _is_hidden_config(loc: Path) -> bool:
    """
    Want to keep rules/.semgrep.yml but not path/.github/foo.yml
    Also want to keep src/.semgrep/bad_pattern.yml but not ./.pre-commit-config.yaml
    """
    return any(
        part != "."
        and part != ".."
        and part.startswith(".")
        and DEFAULT_SEMGREP_CONFIG_NAME not in part
        for part in loc.parts
    )


def load_config_from_local_path(
    location: Optional[str] = None,
) -> Dict[str, Optional[Dict[str, Any]]]:
    """
        Return config file(s) as dictionary object
    """
    base_path = get_base_path()
    if location is None:
        default_file = base_path.joinpath(DEFAULT_CONFIG_FILE)
        default_folder = base_path.joinpath(DEFAULT_CONFIG_FOLDER)
        if default_file.exists():
            return parse_config_at_path(default_file)
        elif default_folder.exists():
            return parse_config_folder(default_folder, relative=True)
        else:
            return {}
    else:
        loc = base_path.joinpath(location)
        if loc.exists():
            if loc.is_file():
                return parse_config_at_path(loc)
            elif loc.is_dir():
                return parse_config_folder(loc)
            else:
                print_error_exit(f"config location `{loc}` is not a file or folder!")
        else:
            addendum = ""
            if IN_DOCKER:
                addendum = " (since you are running in docker, you cannot specify arbitary paths on the host; they must be mounted into the container)"
            print_error_exit(
                f"unable to find a config; path `{loc}` does not exist{addendum}"
            )
    raise Exception


def download_config(config_url: str) -> Dict[str, Optional[Dict[str, Any]]]:
    import requests  # here for faster startup times

    DOWNLOADING_MESSAGE = "downloading config..."
    SCANNING_MESSAGE = "scanning code...     "  # needs to have same number of chars as DOWNLOADING_MESSAGE
    debug_print(f"trying to download from {config_url}")
    print_msg(DOWNLOADING_MESSAGE, end="\r")
    headers = {"User-Agent": SEMGREP_USER_AGENT}

    try:
        r = requests.get(config_url, stream=True, headers=headers, timeout=10)
        if r.status_code == requests.codes.ok:
            content_type = r.headers.get("Content-Type")
            if content_type and "text/plain" in content_type:
                print_msg(SCANNING_MESSAGE)
                return parse_config_string("remote-url", r.content.decode("utf-8"))
            elif content_type and content_type == "application/x-gzip":
                with tempfile.TemporaryDirectory() as fname:
                    with tarfile.open(fileobj=r.raw, mode="r:gz") as tar:
                        tar.extractall(fname)
                    extracted = Path(fname)
                    for path in extracted.iterdir():
                        # get first folder in extracted folder (this is how GH does it)
                        print_msg(SCANNING_MESSAGE)
                        return parse_config_folder(path, relative=True)
            else:
                print_error_exit(
                    f"unknown content-type: {content_type} returned by config url: {config_url}. Can not parse"
                )
        else:
            print_error_exit(
                f"bad status code: {r.status_code} returned by config url: {config_url}"
            )
    except Exception as e:
        print_error(str(e))
    return {config_url: None}


def resolve_config(config_str: Optional[str]) -> Dict[str, Optional[Dict[str, Any]]]:
    """ resolves if config arg is a registry entry, a url, or a file, folder, or loads from defaults if None"""
    start_t = time.time()
    if config_str is None:
        config = load_config_from_local_path()
    elif config_str in RULES_REGISTRY:
        config = download_config(RULES_REGISTRY[config_str])
    elif is_url(config_str):
        config = download_config(config_str)
    else:
        config = load_config_from_local_path(config_str)
    if config:
        debug_print(f"loaded {len(config)} configs in {time.time() - start_t}")
    return config


def generate_config() -> None:
    import requests  # here for faster startup times

    # defensive coding
    if Path(DEFAULT_CONFIG_FILE).exists():
        print_error_exit(
            f"{DEFAULT_CONFIG_FILE} already exists. Please remove and try again"
        )
    try:
        r = requests.get(TEMPLATE_YAML_URL, timeout=10)
        r.raise_for_status()
        template_str = r.text
    except Exception as e:
        debug_print(str(e))
        print_msg(
            f"There was a problem downloading the latest template config. Using fallback template"
        )
        template_str = """rules:
  - id: eqeq-is-bad
    pattern: $X == $X
    message: "$X == $X is a useless equality check"
    languages: [python]
    severity: ERROR"""
    try:
        with open(DEFAULT_CONFIG_FILE, "w") as template:
            template.write(template_str)
            print_msg(f"Template config successfully written to {DEFAULT_CONFIG_FILE}")
            sys.exit(0)
    except Exception as e:
        print_error_exit(str(e))



import itertools
import sys
from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Tuple
from urllib.parse import urlparse

global DEBUG
global QUIET
DEBUG = False
QUIET = False

# Exit codes
FINDINGS_EXIT_CODE = 1
FATAL_EXIT_CODE = 2
INVALID_CODE_EXIT_CODE = 3
INVALID_PATTERN_EXIT_CODE = 4
UNPARSEABLE_YAML_EXIT_CODE = 5
NEED_ARBITRARY_CODE_EXEC_EXIT_CODE = 6
MISSING_CONFIG_EXIT_CODE = 7


def is_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def print_error(e: str) -> None:
    if not QUIET:
        print(e, file=sys.stderr)


def print_error_exit(msg: str, exit_code: int = FATAL_EXIT_CODE) -> None:
    if not QUIET:
        print(msg, file=sys.stderr)
    sys.exit(exit_code)


def print_msg(msg: str, **kwargs: Any) -> None:
    if not QUIET:
        print(msg, file=sys.stderr, **kwargs)


def debug_print(msg: str) -> None:
    if DEBUG:
        print(msg, file=sys.stderr)


def flatten(L: Iterable[Iterable[Any]]) -> Iterable[Any]:
    for list in L:
        for item in list:
            yield item


def set_flags(debug: bool, quiet: bool) -> None:
    """Set the global DEBUG and QUIET flags"""
    # TODO move to a proper logging framework
    global DEBUG
    global QUIET
    if debug:
        DEBUG = True
        debug_print("DEBUG is on")
    if quiet:
        QUIET = True
        debug_print("QUIET is on")


def partition(pred: Callable, iterable: Iterable) -> Tuple[List, List]:
    """E.g. partition(is_odd, range(10)) -> 1 3 5 7 9  and  0 2 4 6 8"""
    i1, i2 = itertools.tee(iterable)
    return list(filter(pred, i1)), list(itertools.filterfalse(pred, i2))



import subprocess
from pathlib import Path
from typing import NamedTuple

import pytest


class RepoCase(NamedTuple):
    url: str
    sha: str
    language: str

    @property
    def test_id(self):
        return self.url.split("/")[-1]


@pytest.mark.parametrize(
    "repo_case",
    [
        # SHAs of May 6, 2020
        RepoCase("https://github.com/getsentry/sentry", "f25ea5dc", "python"),
        RepoCase("https://github.com/highcharts/highcharts", "04709689", "javascript"),
    ],
    ids=lambda case: case.test_id,
)
def test_public_repos(run_semgrep_in_tmp, benchmark, repo_case):
    subprocess.check_output(["git", "clone", repo_case.url, "repo"])
    subprocess.check_output(
        ["git", "--git-dir", "repo/.git", "checkout", repo_case.sha]
    )
    benchmark(
        subprocess.check_output,
        [
            "python3",
            "-m",
            "semgrep",
            "--jobs",
            "1",
            "--pattern",
            "$X = 156128578192758",
            "--lang",
            repo_case.language,
            "repo",
        ],
    )



import subprocess
from pathlib import Path

import pytest

# output of `fd ya?ml$`
# not calculating dynamically to avoid slow test collection on each pytest run
RULE_PATHS: str = """
c/lang/correctness/goto-fail.yaml
contrib/nodejsscan/archive_path_overwrite.yaml
contrib/nodejsscan/crypto_node.yaml
contrib/nodejsscan/directory_traversal.yaml
contrib/nodejsscan/error_disclosure.yaml
contrib/nodejsscan/eval_deserialize.yaml
contrib/nodejsscan/eval_node.yaml
contrib/nodejsscan/eval_yaml_deserialize.yaml
contrib/nodejsscan/exec_os_command.yaml
contrib/nodejsscan/express_bodyparser_dos.yaml
contrib/nodejsscan/hardcoded_jwt.yaml
contrib/nodejsscan/hardcoded_secrets.yaml
contrib/nodejsscan/header_cors_star.yaml
contrib/nodejsscan/header_helmet_disabled.yaml
contrib/nodejsscan/header_injection.yaml
contrib/nodejsscan/header_xss_protection.yaml
contrib/nodejsscan/jwt_none_algorithm.yaml
contrib/nodejsscan/logic_bypass.yaml
contrib/nodejsscan/nosql_injection.yaml
contrib/nodejsscan/open_redirect.yaml
contrib/nodejsscan/regex_injection.yaml
contrib/nodejsscan/security_electronjs.yaml
contrib/nodejsscan/server_side_template_injection.yaml
contrib/nodejsscan/sql_injection.yaml
contrib/nodejsscan/ssrf_node.yaml
contrib/nodejsscan/timing_attack_node.yaml
contrib/nodejsscan/tls_node.yaml
contrib/nodejsscan/xml_entity_expansion_dos.yaml
contrib/nodejsscan/xpathi_node.yaml
contrib/nodejsscan/xss_node.yaml
contrib/nodejsscan/xss_templates.yaml
contrib/nodejsscan/xxe_node.yaml
go/gorilla/security/audit/handler-assignment-from-multiple-sources.yaml
go/gorilla/security/audit/handler-attribute-read-from-multiple-sources.yaml
go/jwt-go/security/jwt.yaml
go/lang/best-practice/channel-guarded-with-mutex.yaml
go/lang/best-practice/hidden-goroutine.yaml
go/lang/correctness/bad_temp_file/bad_tmp.yaml
go/lang/correctness/dos/decompression_bomb.yaml
go/lang/correctness/overflow/overflow.yaml
go/lang/correctness/permissions/file_permission.yaml
go/lang/correctness/traversal/zip.yaml
go/lang/correctness/unsafe.yaml
go/lang/correctness/useless-eqeq.yaml
go/lang/maintainability/useless-ifelse.yaml
go/lang/security/audit/crypto/bad_imports.yaml
go/lang/security/audit/crypto/insecure_ssh.yaml
go/lang/security/audit/crypto/math_random.yaml
go/lang/security/audit/crypto/ssl.yaml
go/lang/security/audit/crypto/tls.yaml
go/lang/security/audit/crypto/use_of_weak_crypto.yaml
go/lang/security/audit/crypto/use_of_weak_rsa_key.yaml
go/lang/security/audit/database/string-formatted-query.yaml
go/lang/security/audit/net/bind_all.yaml
go/lang/security/audit/net/formatted-template-string.yaml
go/lang/security/audit/net/pprof.yaml
go/lang/security/audit/net/unescaped-data-in-htmlattr.yaml
go/lang/security/audit/net/unescaped-data-in-js.yaml
go/lang/security/audit/net/unescaped-data-in-url.yaml
go/lang/security/audit/net/use-tls.yaml
go/lang/security/audit/net/wip-xss-using-responsewriter-and-printf.yaml
java/lang/correctness/assignment-comparison.yaml
java/lang/correctness/eqeq.yaml
java/lang/correctness/hardcoded-conditional.yaml
java/lang/security/audit/anonymous-ldap-bind.yaml
java/lang/security/audit/bad-hexa-conversion.yaml
java/lang/security/audit/command-injection-formatted-runtime-call.yaml
java/lang/security/audit/crlf-injection-logs.yaml
java/lang/security/audit/crypto/desede-is-deprecated.yaml
java/lang/security/audit/crypto/ecb-cipher.yaml
java/lang/security/audit/crypto/no-null-cipher.yaml
java/lang/security/audit/crypto/no-static-initialization-vector.yaml
java/lang/security/audit/crypto/rsa-no-padding.yaml
java/lang/security/audit/crypto/ssl/avoid-implementing-custom-digests.yaml
java/lang/security/audit/crypto/ssl/defaulthttpclient-is-deprecated.yaml
java/lang/security/audit/crypto/ssl/insecure-hostname-verifier.yaml
java/lang/security/audit/crypto/ssl/insecure-trust-manager.yaml
java/lang/security/audit/crypto/unencrypted-socket.yaml
java/lang/security/audit/crypto/weak-hash.yaml
java/lang/security/audit/crypto/weak-rsa.yaml
java/lang/security/audit/el-injection.yaml
java/lang/security/audit/formatted-sql-string.yaml
java/lang/security/audit/jdbc-sql-formatted-string.yaml
java/lang/security/audit/ldap-entry-poisoning.yaml
java/lang/security/audit/ldap-injection.yaml
java/lang/security/audit/script-engine-injection.yaml
java/lang/security/audit/unvalidated-redirect.yaml
java/lang/security/audit/url-rewriting.yaml
java/lang/security/audit/weak-ssl-context.yaml
java/lang/security/audit/xml-decoder.yaml
java/lang/security/servletresponse-writer-xss.yaml
java/spring/security/audit/spel-injection.yaml
java/spring/security/audit/spring-cookie-missing-httponly.yaml
java/spring/security/audit/spring-cookie-missing-secure-flag.yaml
java/spring/security/audit/spring-csrf-disabled.yaml
java/spring/security/audit/spring-unvalidated-redirect.yaml
javascript/jose/security/jwt-exposed-credentials.yaml
javascript/jose/security/jwt-hardcode.yaml
javascript/jose/security/jwt-none-alg.yaml
javascript/jsonwebtoken/security/audit/jwt-decode-without-verify.yaml
javascript/jsonwebtoken/security/jwt-exposed-credentials.yaml
javascript/jsonwebtoken/security/jwt-hardcode.yaml
javascript/jsonwebtoken/security/jwt-none-alg.yaml
javascript/lang/best-practice/assigned-undefined.yaml
javascript/lang/best-practice/leftover_debugging.yaml
javascript/lang/correctness/useless-assign.yaml
javascript/lang/correctness/useless-eqeq.yaml
javascript/lang/security/path-traversal/path-join-resolve-traversal.yaml
javascript/lang/security/spawn-git-clone.yaml
javascript/passport-jwt/security/passport-hardcode.yaml
python/airflow/security/audit/formatted-string-bashoperator.yaml
python/bokeh/maintainability/deprecated/deprecated_apis.yaml
python/boto3/security/hardcoded-token.yaml
python/click/best-practice/echo-style.yaml
python/django/best-practice/json_response.yaml
python/django/best-practice/upsell_django_environ.yaml
python/django/best-practice/use-onetoonefield.yaml
python/django/compatibility/django-2_0-compat.yaml
python/django/correctness/model-save.yaml
python/django/correctness/nontext-field-must-set-null-true.yaml
python/django/correctness/string-field-null-checks.yaml
python/django/correctness/use-decimalfield-for-money.yaml
python/django/performance/access-foreign-keys.yaml
python/django/performance/upsell-count.yaml
python/django/performance/upsell_earliest_latest.yaml
python/django/security/audit/csrf-exempt.yaml
python/django/security/audit/custom-expression-as-sql.yaml
python/django/security/audit/extends-custom-expression.yaml
python/django/security/audit/query-set-extra.yaml
python/django/security/audit/raw-query.yaml
python/django/security/audit/secure-cookies.yaml
python/django/security/audit/unvalidated-password.yaml
python/django/security/injection/code/globals-misuse-code-execution.yaml
python/django/security/injection/code/user-eval-format-string.yaml
python/django/security/injection/code/user-eval.yaml
python/django/security/injection/code/user-exec-format-string.yaml
python/django/security/injection/code/user-exec.yaml
python/django/security/injection/command/command-injection-os-system.yaml
python/django/security/injection/email/xss-html-email-body.yaml
python/django/security/injection/email/xss-send-mail-html-message.yaml
python/django/security/injection/mass-assignment.yaml
python/django/security/injection/open-redirect.yaml
python/django/security/injection/path-traversal/path-traversal-file-name.yaml
python/django/security/injection/path-traversal/path-traversal-join.yaml
python/django/security/injection/path-traversal/path-traversal-open.yaml
python/django/security/injection/reflected-data-httpresponse.yaml
python/django/security/injection/reflected-data-httpresponsebadrequest.yaml
python/django/security/injection/request-data-fileresponse.yaml
python/django/security/injection/request-data-write.yaml
python/django/security/injection/sql/sql-injection-extra.yaml
python/django/security/injection/sql/sql-injection-rawsql.yaml
python/django/security/injection/sql/sql-injection-using-db-cursor-execute.yaml
python/django/security/injection/sql/sql-injection-using-raw.yaml
python/django/security/injection/ssrf/ssrf-injection-requests.yaml
python/django/security/injection/ssrf/ssrf-injection-urllib.yaml
python/django/security/passwords/password-empty-string.yaml
python/django/security/passwords/use-none-for-password-default.yaml
python/flask/best-practice/use-jsonify.yaml
python/flask/correctness/access-request-in-wrong-handler.yaml
python/flask/correctness/same-handler-name.yaml
python/flask/experimental/correctness/different-route-names.yaml
python/flask/experimental/security/audit/secure-cookies/secure-session-cookies.yaml
python/flask/maintainability/deprecated/deprecated-apis.yaml
python/flask/security/audit/app-run-param-config.yaml
python/flask/security/audit/app-run-security-config.yaml
python/flask/security/audit/hardcoded-config.yaml
python/flask/security/audit/secure-set-cookie.yaml
python/flask/security/audit/wtf-csrf-disabled.yaml
python/flask/security/open-redirect.yaml
python/flask/security/render-template-string.yaml
python/flask/security/secure-static-file-serve.yaml
python/flask/security/unescaped-template-extension.yaml
python/flask/security/unsanitized_input.yaml
python/lang/best-practice/manual-collections-create.yaml
python/lang/best-practice/missing-hash-with-eq.yaml
python/lang/best-practice/open-never-closed.yaml
python/lang/best-practice/pass-body.yaml
python/lang/best-practice/sleep.yaml
python/lang/compatibility/python36.yaml
python/lang/compatibility/python37.yaml
python/lang/correctness/baseclass-attribute-override.yaml
python/lang/correctness/common-mistakes/default-mutable-dict.yaml
python/lang/correctness/common-mistakes/default-mutable-list.yaml
python/lang/correctness/common-mistakes/is-comparison-string.yaml
python/lang/correctness/common-mistakes/is-not-is-not.yaml
python/lang/correctness/dict-modify-iterating.yaml
python/lang/correctness/exceptions/exceptions.yaml
python/lang/correctness/exit.yaml
python/lang/correctness/list-modify-iterating.yaml
python/lang/correctness/missing-fstring.yaml
python/lang/correctness/return-in-init.yaml
python/lang/correctness/tempfile/flush.yaml
python/lang/correctness/tempfile/mktemp.yaml
python/lang/correctness/unchecked-returns.yaml
python/lang/correctness/useless-comparison.yaml
python/lang/correctness/useless-eqeq.yaml
python/lang/maintainability/return.yaml
python/lang/maintainability/useless-assign-keyed.yaml
python/lang/maintainability/useless-assign.notyaml
python/lang/maintainability/useless-ifelse.yaml
python/lang/maintainability/useless-innerfunction.yaml
python/lang/maintainability/useless-literal.yaml
python/lang/security/audit/conn_recv.yaml
python/lang/security/audit/logging/listeneval.yaml
python/lang/security/audit/marshal.yaml
python/lang/security/audit/network/bind.yaml
python/lang/security/audit/network/disabled-cert-validation.yaml
python/requests/best-practice/use-timeout.yaml
python/requests/security/no-auth-over-http.yaml
python/sqlalchemy/correctness/delete-where.yaml
python/sqlalchemy/performance/performance-improvements.yaml
"""


@pytest.fixture(scope="session")
def semgrep_rules_repo(request, tmp_path_factory):
    repo_path = tmp_path_factory.mktemp("repo")

    getoption = request.config.getoption
    is_disabled = getoption("benchmark_skip") or getoption("benchmark_disable")
    is_force_enabled = getoption("benchmark_enable") or getoption("benchmark_only")
    if is_disabled and not is_force_enabled:
        return repo_path  # not cloning since we're not gonna run benchmarks

    subprocess.check_output(
        ["git", "clone", "https://github.com/returntocorp/semgrep-rules", repo_path]
    )
    subprocess.check_output(
        ["git", "checkout", "c3196b4"], cwd=repo_path  # May 16, 2020
    )

    return repo_path


@pytest.fixture(params=RULE_PATHS.strip().splitlines())
def semgrep_rules_rule(semgrep_rules_repo, request, tmp_path, benchmark):
    rule_path = semgrep_rules_repo / Path(request.param)

    # we gather the rule and any non-rule files, which each might be unit tests
    needed_paths = [
        child
        for child in rule_path.parent.iterdir()
        if child.suffix not in {".yml", ".yaml"} or child.name == rule_path.name
    ]
    for needed_path in needed_paths:
        (tmp_path / needed_path.name).symlink_to(needed_path.resolve())
    yield tmp_path


def test_semgrep_rules_rule(semgrep_rules_rule, benchmark):
    benchmark(
        subprocess.check_output,
        [
            "python3",
            "-m",
            "semgrep",
            "--jobs",
            "1",
            "--dangerously-allow-arbitrary-code-execution-from-rules",
            "--strict",
            "--test",
            "--test-ignore-todo",
            semgrep_rules_rule,
        ],
    )



def foo(request):
    # ERROR: open-from-request
    path = request.get("unsafe")
    open(path)

def bar(request):
    foo = request.get("unsafe")
    path = "safe_path"
    open(path)

# This _should_ be detected, but is not
# due to semgrep#707
z = request.get("unsafe")
def baz():
    open(z)




from typing import Any
from typing import Dict
from typing import List

from semgrep.semgrep_types import BooleanRuleExpression


class Pattern:
    """
        Subunit of Rule that semgrep-core runs on
    """

    def __init__(
        self,
        rule_index: int,
        expression: BooleanRuleExpression,
        severity: str,
        languages: List[str],
    ) -> None:
        self._id = f"{rule_index}.{expression.pattern_id}"
        # if we don't copy an array (like `languages`), the yaml file will refer to it by reference (with an anchor)
        # which is nice and all but the semgrep YAML parser doesn't support that
        self._languages = languages.copy()
        self._severity = severity
        self._expression = expression
        self._pattern = expression.operand
        self._message = "<internalonly>"

    @property
    def languages(self) -> List[str]:
        return self._languages

    @property
    def expression(self) -> BooleanRuleExpression:
        return self._expression

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self._id,
            "pattern": self._pattern,
            "severity": self._severity,
            "languages": self._languages,
            "message": self._message,
        }



import json
from pathlib import Path
from subprocess import CalledProcessError

import pytest


def test_basic_rule__local(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(run_semgrep_in_tmp("rules/eqeq.yaml"), "results.json")


def test_basic_rule__relative(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/../rules/eqeq.yaml"), "results.json",
    )


def test_basic_rule__absolute(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(Path.cwd() / "rules" / "eqeq.yaml"), "results.json",
    )


def test_terminal_output(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("rules/eqeq.yaml", output_format="normal"), "output.txt"
    )


def test_sarif_output(run_semgrep_in_tmp, snapshot):
    sarif_output = json.loads(
        run_semgrep_in_tmp("rules/eqeq.yaml", output_format="sarif")
    )

    # rules are logically a set so the JSON list's order doesn't matter
    # we make the order deterministic here so that snapshots match across runs
    # the proper solution will be https://github.com/joseph-roitman/pytest-snapshot/issues/14
    sarif_output["tool"]["driver"]["rules"] = sorted(
        sarif_output["tool"]["driver"]["rules"], key=lambda rule: rule["id"]
    )

    snapshot.assert_match(
        json.dumps(sarif_output, indent=2, sort_keys=True), "results.sarif"
    )


def test_url_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp(
            "https://raw.githubusercontent.com/returntocorp/semgrep-rules/develop/template.yaml",
        ),
        "results.json",
    )


def test_registry_rule(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(
        run_semgrep_in_tmp("r2c"), "results.json",
    )


def test_hidden_rule__explicit(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(run_semgrep_in_tmp("rules/hidden/.hidden"), "results.json")


def test_hidden_rule__implicit(run_semgrep_in_tmp, snapshot):
    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp("rules/hidden", stderr=True)
    assert excinfo.value.returncode == 2
    snapshot.assert_match(excinfo.value.output, "error.txt")


def test_default_rule__file(run_semgrep_in_tmp, snapshot):
    Path(".semgrep.yml").symlink_to(Path("rules/eqeq.yaml").resolve())
    snapshot.assert_match(run_semgrep_in_tmp(), "results.json")


def test_default_rule__folder(run_semgrep_in_tmp, snapshot):
    Path(".semgrep").mkdir()
    Path(".semgrep/.semgrep.yml").symlink_to(Path("rules/eqeq.yaml").resolve())

    snapshot.assert_match(run_semgrep_in_tmp(), "results.json")


def test_regex_rule__top(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(run_semgrep_in_tmp("rules/regex-top.yaml"), "results.json")


def test_regex_rule__child(run_semgrep_in_tmp, snapshot):
    snapshot.assert_match(run_semgrep_in_tmp("rules/regex-child.yaml"), "results.json")


def test_regex_rule__invalid_expression(run_semgrep_in_tmp, snapshot):
    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp("rules/regex-invalid.yaml", stderr=True)
    assert excinfo.value.returncode == 2
    snapshot.assert_match(excinfo.value.output, "error.txt")



from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.util import print_error_exit
from semgrep.util import print_msg


def _modify_file(rule_match: RuleMatch, fix: str) -> None:
    p = Path(rule_match.path)
    SPLIT_CHAR = "\n"
    contents = p.read_text()
    lines = contents.split(SPLIT_CHAR)

    # get the start and end points
    start_obj = rule_match.start
    start_line = start_obj.get("line", 1) - 1  # start_line is 1 indexed
    start_col = start_obj.get("col", 1) - 1  # start_col is 1 indexed
    end_obj = rule_match.end
    end_line = end_obj.get("line", 1) - 1  # end_line is 1 indexed
    end_col = end_obj.get("col", 1) - 1  # end_line is 1 indexed

    # break into before, to modify, after
    before_lines = lines[:start_line]
    before_on_start_line = lines[start_line][:start_col]
    after_on_end_line = lines[end_line][end_col + 1 :]  # next char after end of match
    modified_lines = (before_on_start_line + fix + after_on_end_line).splitlines()
    after_lines = lines[end_line + 1 :]  # next line after end of match
    contents_after_fix = before_lines + modified_lines + after_lines

    contents_after_fix_str = SPLIT_CHAR.join(contents_after_fix)
    p.write_text(contents_after_fix_str)


def apply_fixes(rule_matches_by_rule: Dict[Rule, List[RuleMatch]]) -> None:
    """
        Modify files in place for all files with findings from rules with an
        autofix configuration
    """
    modified_files: Set[Path] = set()

    for rule, rule_matches in rule_matches_by_rule.items():
        for rule_match in rule_matches:
            fix = rule_match.fix
            if fix:
                filepath = rule_match.path
                try:
                    _modify_file(rule_match, fix)
                    modified_files.add(filepath)
                except Exception as e:
                    print_error_exit(f"unable to modify file: {filepath}: {e}")
    num_modified = len(modified_files)
    print_msg(
        f"Successfully modified {num_modified} file{'s' if num_modified > 1 else ''}."
    )



import os
import subprocess
import sys
from enum import auto
from enum import Enum

RCE_RULE_FLAG = "--dangerously-allow-arbitrary-code-execution-from-rules"
RULES_KEY = "rules"
SEMGREP_USER_AGENT = "semgrep-give-me-yaml"
ID_KEY = "id"
SEMGREP_URL = "https://semgrep.dev/"
PLEASE_FILE_ISSUE_TEXT = "An error occurred while invoking the semgrep engine; please help us fix this by filing an an issue at https://semgrep.dev"

DEFAULT_SEMGREP_CONFIG_NAME = "semgrep"
DEFAULT_CONFIG_FILE = f".{DEFAULT_SEMGREP_CONFIG_NAME}.yml"
DEFAULT_CONFIG_FOLDER = f".{DEFAULT_SEMGREP_CONFIG_NAME}"

YML_EXTENSIONS = {".yml", ".yaml"}

try:
    from semgrep.version import version

    __VERSION__ = version
except ModuleNotFoundError:
    __VERSION__ = "local"


def compute_semgrep_path() -> str:
    exec_name = "semgrep-core"
    if subprocess.run(["which", exec_name], stdout=subprocess.DEVNULL).returncode != 0:
        # look for something in the same dir as the Python interpreter
        relative_path = os.path.join(os.path.dirname(sys.executable), exec_name)
        if os.path.exists(relative_path):
            exec_name = relative_path
    return exec_name


SEMGREP_PATH = compute_semgrep_path()


class OutputFormat(Enum):
    TEXT = auto()
    JSON = auto()
    SARIF = auto()



import json
import subprocess
from pathlib import Path

import pytest

pytestmark = pytest.mark.qa
SENTINEL_VALUE = 87518275812375164


def xfail_repo(url, *, reason=None):
    return pytest.param(url, marks=pytest.mark.xfail(reason=reason))


@pytest.mark.parametrize(
    "repo_url",
    [
        "https://github.com/Airtable/airtable.js",
        "https://github.com/seemoo-lab/opendrop",
        "https://github.com/lightstep/lightstep-tracer-python",
        "https://github.com/draios/sysdig-inspect",
        "https://github.com/getsentry/sentry-python",
        "https://github.com/signalapp/signal-webrtc-ios",
        "https://github.com/secdev/scapy",
        "https://github.com/apache/airflow",
        "https://github.com/preset-io/elasticsearch-dbapi",
        "https://github.com/apache/libcloud",
        "https://github.com/keybase/pykeybasebot",
        "https://gitbox.apache.org/repos/asf/cassandra",
        "https://github.com/coinbase/coinbase-commerce-python",
        "https://github.com/keybase/python-triplesec",
        "https://github.com/psycopg/psycopg2",
        "https://github.com/preset-io/flask-jwt-extended",
        "https://github.com/vstinner/pyperf",
        "https://github.com/mysql/mysql-connector-python",
        "https://github.com/Netflix/lemur",
        xfail_repo("https://github.com/highcharts/highcharts"),
        xfail_repo(
            "https://github.com/lodash/lodash",
            reason="https://github.com/returntocorp/semgrep/issues/580",
        ),
        xfail_repo("https://github.com/signalapp/Signal-Desktop"),
        xfail_repo("https://github.com/opensourceactivismtech/call-power"),
        xfail_repo("https://github.com/zulip/zulip"),
        xfail_repo(
            "https://github.com/home-assistant/home-assistant",
            reason=(
                "https://github.com/returntocorp/semgrep/issues/599, "
                "https://github.com/returntocorp/semgrep/issues/600, "
                "https://github.com/returntocorp/semgrep/issues/601, "
                "https://github.com/returntocorp/semgrep/issues/602"
            ),
        ),
        xfail_repo(
            "https://github.com/apache/incubator-superset",
            reason=(
                "https://github.com/returntocorp/semgrep/issues/581, "
                "https://github.com/returntocorp/semgrep/issues/582"
            ),
        ),
    ],
)
def test_semgrep_on_repo(monkeypatch, tmp_path, repo_url):
    TESTS_PATH = Path(__file__).parent.parent
    monkeypatch.setenv("PYTHONPATH", str(TESTS_PATH.parent.resolve()))
    (tmp_path / "rules").symlink_to(Path(TESTS_PATH / "qa" / "rules").resolve())
    monkeypatch.chdir(tmp_path)

    subprocess.check_output(
        [
            "git",
            "clone",
            "--depth=1",
            "https://github.com/returntocorp/semgrep-rules",
            "repo",
        ]
    )
    languages = {
        "python": "py",
        "javascript": "js",
    }
    for language, file_ext in languages.items():
        sentinel_path = Path("repo") / f"sentinel.{file_ext}"
        with sentinel_path.open("w") as sentinel_file:
            sentinel_file.write(f"x = {SENTINEL_VALUE}")

        semgrep_run = subprocess.run(
            [
                "python3",
                "-m",
                "semgrep",
                "--pattern",
                f"$X = {SENTINEL_VALUE}",
                "--lang",
                language,
                "--json",
                "repo",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )

        assert semgrep_run.returncode == 0
        try:
            output = json.loads(semgrep_run.stdout)
        except json.JSONDecodeError:
            pytest.fail(
                f"Failed to parse JSON from semgrep output:\n"
                + semgrep_run.stdout
                + semgrep_run.stderr
            )

        assert output["errors"] == []
        assert len(output["results"]) == 1
        assert output["results"][0]["path"] == str(sentinel_path)

    output = subprocess.check_output(
        [
            "python3",
            "-m",
            "semgrep",
            "--config=rules/regex-sentinel.yaml",
            "--strict",
            "--json",
            "repo",
        ],
        encoding="utf-8",
    )
    output = json.loads(output)
    assert len(output["results"]) == 2
    assert len(output["errors"]) == 0



#!/usr/bin/env python3
import argparse
import multiprocessing
import os
import sys

import semgrep.config_resolver
import semgrep.semgrep_main
import semgrep.test
from semgrep.constants import __VERSION__
from semgrep.constants import DEFAULT_CONFIG_FILE
from semgrep.constants import PLEASE_FILE_ISSUE_TEXT
from semgrep.constants import RCE_RULE_FLAG
from semgrep.constants import SEMGREP_URL
from semgrep.dump_ast import dump_parsed_ast
from semgrep.util import print_error
from semgrep.util import print_error_exit


try:
    CPU_COUNT = multiprocessing.cpu_count()
except NotImplementedError:
    CPU_COUNT = 1  # CPU count is not implemented on Windows


def cli() -> None:
    parser = argparse.ArgumentParser(
        description=f"semgrep CLI. For more information about semgrep, go to {SEMGREP_URL}",
        prog="semgrep",
    )

    # input
    parser.add_argument(
        "target",
        nargs="*",
        default=[os.curdir],
        help=(
            "Search these files or directories. Defaults to entire current "
            "working directory. Implied argument if piping to semgrep."
        ),
    )

    # config options
    config = parser.add_argument_group("config")
    config_ex = config.add_mutually_exclusive_group()
    config_ex.add_argument(
        "-g",
        "--generate-config",
        action="store_true",
        help=f"Generate starter configuration file, {DEFAULT_CONFIG_FILE}",
    )

    config_ex.add_argument(
        "-f",
        "--config",
        help=(
            "YAML configuration file, directory of YAML files ending in "
            ".yml|.yaml, URL of a configuration file, or semgrep registry entry "
            "name. See README for information on configuration file format."
        ),
    )

    config_ex.add_argument(
        "-e",
        "--pattern",
        help="Code search pattern. See README for information on pattern features.",
    )
    config.add_argument(
        "-l",
        "--lang",
        help=(
            "Parse pattern and all files in specified language. Must be used "
            "with -e/--pattern."
        ),
    )
    config.add_argument(
        "--validate",
        action="store_true",
        help="Validate configuration file(s). No search is performed.",
    )
    config.add_argument(
        "--strict",
        action="store_true",
        help="Only invoke semgrep if configuration files(s) are valid.",
    )

    parser.add_argument(
        "--exclude",
        action="append",
        default=[],
        help="Skip any file with this name; --exclude='*.py' will ignore foo.py as well as src/foo.py. Can add multiple times. Overrides includes.",
    )
    parser.add_argument(
        "--include",
        action="append",
        default=[],
        help="Scan only files with this name, such as --include='*.jsx'. Can add multiple times.",
    )
    parser.add_argument(
        "--exclude-dir",
        action="append",
        default=[],
        help="Skip any directory with this name; --exclude-dir=doc will ignore doc/ as well as src/doc. Can add multiple times. Overrides includes.",
    )
    parser.add_argument(
        "--include-dir",
        action="append",
        default=[],
        help="Scan only directories with this name; --include-dir=doc will include doc/ as well as src/doc. Can add multiple times.",
    )

    config.add_argument(
        RCE_RULE_FLAG,
        action="store_true",
        help=(
            "WARNING: allow rules to run arbitrary code. ONLY ENABLE IF YOU "
            "TRUST THE SOURCE OF ALL RULES IN YOUR CONFIGURATION."
        ),
    )

    config.add_argument(
        "--exclude-tests",
        action="store_true",
        help="Exclude tests, documentation, and examples based on filename/path.",
    )
    config.add_argument(
        "--precommit", action="store_true", help=argparse.SUPPRESS,
    )
    config.add_argument(
        "-j",
        "--jobs",
        action="store",
        type=int,
        default=CPU_COUNT,
        help=(
            "Number of subprocesses to use to run checks in parallel. Defaults "
            "to the number of CPUs on the system."
        ),
    )

    # output options
    output = parser.add_argument_group("output")

    output.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help=(
            "Do not print anything to stdout. Search results can still be "
            "saved to an output file specified by -o/--output. Exit code "
            "provides success status."
        ),
    )

    output.add_argument(
        "--no-rewrite-rule-ids",
        action="store_true",
        help=(
            "Do not rewrite rule ids when they appear in nested sub-directories "
            "(by default, rule 'foo' in test/rules.yaml will be renamed "
            "'test.foo')."
        ),
    )

    output.add_argument(
        "-o",
        "--output",
        help=(
            "Save search results to a file or post to URL. "
            "Default is to print to stdout."
        ),
    )
    output.add_argument(
        "--json", action="store_true", help="Output results in JSON format."
    )
    output.add_argument(
        "--sarif", action="store_true", help="Output results in SARIF format."
    )
    output.add_argument("--test", action="store_true", help="Run test suite.")
    parser.add_argument(
        "--test-ignore-todo",
        action="store_true",
        help="Ignore rules marked as '#todoruleid:' in test files.",
    )
    output.add_argument(
        "--r2c",
        action="store_true",
        help="Output json in r2c platform format (https://app.r2c.dev).",
    )
    output.add_argument(
        "--dump-ast",
        action="store_true",
        help=(
            "Show AST of the input file or passed expression and then exit "
            "(can use --json)."
        ),
    )
    output.add_argument(
        "--error",
        action="store_true",
        help="Exit 1 if there are findings. Useful for CI and scripts.",
    )

    output.add_argument(
        "-a",
        "--autofix",
        action="store_true",
        help=(
            "Apply the autofix patches. WARNING: data loss can occur with this "
            "flag. Make sure your files are stored in a version control system."
        ),
    )

    # logging options
    logging = parser.add_argument_group("logging")

    logging.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=(
            "Set the logging level to verbose. E.g. statements about which "
            "files are being processed will be printed."
        ),
    )

    parser.add_argument(
        "--version", action="store_true", help="Show the version and exit."
    )

    ### Parse and validate
    args = parser.parse_args()
    if args.version:
        print(__VERSION__)
        sys.exit(0)

    if args.pattern and not args.lang:
        parser.error("-e/--pattern and -l/--lang must both be specified")

    # set the flags
    semgrep.util.set_flags(args.verbose, args.quiet)

    # change cwd if using docker
    semgrep.config_resolver.adjust_for_docker(args.precommit)

    try:
        if args.dump_ast:
            if not args.lang:
                print_error_exit("language must be specified to dump ASTs")
            else:
                dump_parsed_ast(args.json, args.lang, args.pattern, args.target)
        elif args.validate:
            _, invalid_configs = semgrep.semgrep_main.get_config(
                args.generate_config, args.pattern, args.lang, args.config
            )
            if invalid_configs:
                print_error_exit(
                    f"run with --validate and there were {len(invalid_configs)} errors loading configs"
                )
            else:
                print_error("Config is valid")

        elif args.test:
            semgrep.test.test_main(args)
        else:
            semgrep.semgrep_main.main(
                target=args.target,
                pattern=args.pattern,
                lang=args.lang,
                config=args.config,
                generate_config=args.generate_config,
                no_rewrite_rule_ids=args.no_rewrite_rule_ids,
                jobs=args.jobs,
                include=args.include,
                include_dir=args.include_dir,
                exclude=args.exclude,
                exclude_dir=args.exclude_dir,
                exclude_tests=args.exclude_tests,
                json_format=args.json,
                sarif=args.sarif,
                output_destination=args.output,
                quiet=args.quiet,
                strict=args.strict,
                exit_on_error=args.error,
                autofix=args.autofix,
                dangerously_allow_arbitrary_code_execution_from_rules=args.dangerously_allow_arbitrary_code_execution_from_rules,
            )
    except NotImplementedError as ex:
        print_error_exit(
            f"semgrep encountered an error: {ex}; this is not your fault. {PLEASE_FILE_ISSUE_TEXT}"
        )



import json
from collections import namedtuple
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional

from semgrep.equivalences import Equivalence
from semgrep.semgrep_types import ALLOWED_GLOB_TYPES
from semgrep.semgrep_types import BooleanRuleExpression
from semgrep.semgrep_types import InvalidRuleSchema
from semgrep.semgrep_types import operator_for_pattern_name
from semgrep.semgrep_types import OPERATORS
from semgrep.semgrep_types import pattern_names_for_operator
from semgrep.semgrep_types import pattern_names_for_operators
from semgrep.semgrep_types import PatternId
from semgrep.semgrep_types import RuleGlobs
from semgrep.semgrep_types import YAML_VALID_TOP_LEVEL_OPERATORS


class Rule:
    def __init__(self, raw: Dict[str, Any]) -> None:
        self._raw = raw
        self._expression = self._build_boolean_expression(raw)
        self._globs = self._build_globs(raw)

    def _parse_boolean_expression(
        self, rule_patterns: List[Dict[str, Any]], pattern_id: int = 0, prefix: str = ""
    ) -> Iterator[BooleanRuleExpression]:
        """
        Move through the expression from the YML, yielding tuples of (operator, unique-id-for-pattern, pattern)
        """
        if not isinstance(rule_patterns, list):
            raise InvalidRuleSchema(
                f"invalid type for patterns in rule: {type(rule_patterns)} is not a list; perhaps your YAML is missing a `-` before {rule_patterns}?"
            )
        for pattern in rule_patterns:
            if not isinstance(pattern, dict):
                raise InvalidRuleSchema(
                    f"invalid type for pattern {pattern}: {type(pattern)} is not a dict"
                )
            for boolean_operator, pattern_text in pattern.items():
                operator = operator_for_pattern_name(boolean_operator)
                if isinstance(pattern_text, list):
                    sub_expression = self._parse_boolean_expression(
                        pattern_text, 0, f"{prefix}.{pattern_id}"
                    )
                    yield BooleanRuleExpression(
                        operator, None, list(sub_expression), None
                    )
                elif isinstance(pattern_text, str):
                    yield BooleanRuleExpression(
                        operator,
                        PatternId(f"{prefix}.{pattern_id}"),
                        None,
                        pattern_text,
                    )
                    pattern_id += 1
                else:
                    raise InvalidRuleSchema(
                        f"invalid type for pattern {pattern}: {type(pattern_text)}"
                    )

    def _build_boolean_expression(
        self, rule_raw: Dict[str, Any]
    ) -> BooleanRuleExpression:
        """
        Build a boolean expression from the yml lines in the rule

        """
        for pattern_name in pattern_names_for_operator(OPERATORS.AND):
            pattern = rule_raw.get(pattern_name)
            if pattern:
                return BooleanRuleExpression(
                    OPERATORS.AND, rule_raw["id"], None, rule_raw[pattern_name]
                )

        for pattern_name in pattern_names_for_operator(OPERATORS.REGEX):
            pattern = rule_raw.get(pattern_name)
            if pattern:
                return BooleanRuleExpression(
                    OPERATORS.REGEX, rule_raw["id"], None, rule_raw[pattern_name]
                )

        for pattern_name in pattern_names_for_operator(OPERATORS.AND_ALL):
            patterns = rule_raw.get(pattern_name)
            if patterns:
                return BooleanRuleExpression(
                    OPERATORS.AND_ALL,
                    None,
                    list(self._parse_boolean_expression(patterns)),
                    None,
                )

        for pattern_name in pattern_names_for_operator(OPERATORS.AND_EITHER):
            patterns = rule_raw.get(pattern_name)
            if patterns:
                return BooleanRuleExpression(
                    OPERATORS.AND_EITHER,
                    None,
                    list(self._parse_boolean_expression(patterns)),
                    None,
                )

        valid_top_level_keys = list(YAML_VALID_TOP_LEVEL_OPERATORS)
        raise InvalidRuleSchema(
            f"missing a pattern type in rule, expected one of {pattern_names_for_operators(valid_top_level_keys)}"
        )

    @staticmethod
    def _build_globs(rule_raw: Dict[str, Any]) -> RuleGlobs:  # type: ignore
        """
        Return a list of globs to be included and excluded for the given `paths:` rules.

        Glob conversion works as follows

        - path: tests/*.py -> tests/*.py
        - directory: tests -> tests/**
        - filename: *.js -> **/*.js
        """
        globs = RuleGlobs(set(), set())

        paths_raw = rule_raw.get("paths", {})
        if not isinstance(paths_raw, dict):
            raise InvalidRuleSchema(
                f"the `paths:` targeting rules must be an object with at least one of {ALLOWED_GLOB_TYPES}"
            )

        for rule_type, rule in rule_raw.get("paths", {}).items():
            if rule_type not in ALLOWED_GLOB_TYPES:
                raise InvalidRuleSchema(
                    f"the `paths:` targeting rules must each be one of {ALLOWED_GLOB_TYPES}"
                )

            glob_set = globs.exclude if rule_type == "exclude" else globs.include
            rule_values = [rule] if isinstance(rule, str) else rule

            for rule_value in rule_values:
                glob_set.add(rule_value)

        return globs

    @property
    def id(self) -> str:
        return str(self._raw["id"])

    @property
    def message(self) -> str:
        return str(self._raw["message"])

    @property
    def metadata(self) -> Dict[str, Any]:  # type: ignore
        return self._raw.get("metadata", {})

    @property
    def severity(self) -> str:
        return str(self._raw["severity"])

    @property
    def sarif_severity(self) -> str:
        """
        SARIF v2.1.0-compliant severity string.

        See https://github.com/oasis-tcs/sarif-spec/blob/a6473580/Schemata/sarif-schema-2.1.0.json#L1566
        """
        mapping = {"INFO": "note", "ERROR": "error", "WARNING": "warning"}
        return mapping[self.severity]

    @property
    def sarif_tags(self) -> Iterator[str]:
        """
        Tags to display on SARIF-compliant UIs, such as GitHub security scans.
        """
        if "cwe" in self.metadata:
            yield "cwe"
        if "owasp" in self.metadata:
            yield "owasp"

    @property
    def languages(self) -> List[str]:
        languages: List[str] = self._raw["languages"]
        return languages

    @property
    def raw(self) -> Dict[str, Any]:  # type: ignore
        return self._raw

    @property
    def expression(self) -> BooleanRuleExpression:
        return self._expression

    @property
    def globs(self) -> RuleGlobs:
        return self._globs

    @property
    def fix(self) -> Optional[str]:
        return self._raw.get("fix")

    @property
    def equivalences(self) -> List[Equivalence]:
        # Use 'i' to make equivalence id's unique
        return [
            Equivalence(f"{self.id}-{i}", eq["equivalence"], self.languages)
            for i, eq in enumerate(self._raw.get(OPERATORS.EQUIVALENCES, []))
        ]

    @classmethod
    def from_json(cls, rule_json: Dict[str, Any]) -> "Rule":  # type: ignore
        return cls(rule_json)

    def to_json(self) -> Dict[str, Any]:
        return self._raw

    def to_sarif(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "name": self.id,
            "shortDescription": {"text": self.message},
            "fullDescription": {"text": self.message},
            "defaultConfiguration": {"level": self.sarif_severity},
            "properties": {"precision": "very-high", "tags": list(self.sarif_tags)},
        }

    def __repr__(self) -> str:
        return json.dumps(self.to_json())



import subprocess


def test_version(benchmark):
    benchmark(subprocess.check_output, ["python3", "-m", "semgrep", "--version"])


def test_eqeq(run_semgrep_in_tmp, benchmark):
    benchmark(run_semgrep_in_tmp, "rules/eqeq.yaml", options=["--jobs", "1"])



# type: ignore
import contextlib
import distutils.util
import os

import setuptools
from setuptools import setup
from setuptools.command.install import install
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel


@contextlib.contextmanager
def chdir(dirname=None):
    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)


# TODO: what is the minimum OSX version?
MIN_OSX_VERSION = "10_14"

# from https://stackoverflow.com/questions/45150304/how-to-force-a-python-wheel-to-be-platform-specific-when-building-it # noqa
class bdist_wheel(_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        # Mark us as not a pure python package (we have platform specific rust code)
        self.root_is_pure = False

    def get_tag(self):
        # this set's us up to build generic wheels.
        # note: we're only doing this for windows right now (causes packaging issues
        # with osx)
        _, _, plat = _bdist_wheel.get_tag(self)
        # to debug "ERROR: *.whl is not a supported wheel on this platform.":
        # from setuptools.pep425tags import get_supported
        # get_supported()
        python = ".".join(["cp36", "cp37", "cp38", "py36", "py37", "py38"])
        abi = "none"

        if "macosx" in plat:
            plat = f"macosx_{MIN_OSX_VERSION}_x86_64.whl"

        # The binary we build is statically linked & manylinux compatible, so change platform
        # accordingly
        if plat == "linux_x86_64":
            plat = "manylinux1_x86_64"
        elif plat == "linux_i686":
            plat = "manylinux1_i686"

        return python, abi, plat


try:
    with open("../README.md") as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = ""

source_dir = os.path.dirname(os.path.abspath(__file__))
repo_root = os.path.dirname(source_dir)

# Lifted with love (and edits) from https://github.com/benfred/py-spy/blob/master/setup.py
class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        if "TOX_ENV_NAME" in os.environ:
            print("Not attempting to install binary while running under tox")
            return
        if "HOMEBREW_SYSTEM" in os.environ:
            print("Not attempting to install binary while running under homebrew")
            install.run(self)
            return
        # So ths builds the executable, and even installs it
        # but we can't install to the bin directory:
        #     https://github.com/pypa/setuptools/issues/210#issuecomment-216657975
        # take the advice from that comment, and move over after install

        if os.environ.get("PRECOMPILED_LOCATION"):
            source = os.environ["PRECOMPILED_LOCATION"]
        else:
            os.putenv("SKIP_NUITKA", "TRUE")
            if "osx" in distutils.util.get_platform():
                with chdir(repo_root):
                    os.system(os.path.join(repo_root, "release-scripts/osx-release.sh"))
                    source = os.path.join(repo_root, "artifacts/semgrep-core")
            else:
                with chdir(repo_root):
                    os.system(
                        os.path.join(repo_root, "release-scripts/ubuntu-release.sh")
                    )
                    source = os.path.join(repo_root, "semgrep-files/semgrep-core")

        ## run this after trying to build (as otherwise this leaves
        ## venv in a bad state: https://github.com/benfred/py-spy/issues/69)
        install.run(self)

        ## we're going to install the semgrep-core executable into the scripts directory
        ## but first make sure the scripts directory exists
        if not os.path.isdir(self.install_scripts):
            os.makedirs(self.install_scripts)

        target = os.path.join(self.install_scripts, "semgrep-core")
        if os.path.isfile(target):
            os.remove(target)

        self.copy_file(source, target)


setup(
    name="semgrep",
    use_scm_version={
        "relative_to": source_dir,
        "write_to": "semgrep/version.py",
        "fallback_version": "0.8.0b1",
    },
    setup_requires=["setuptools_scm", "setuptools_scm_git_archive"],
    author="Return To Corporation",
    author_email="support@r2c.dev",
    description="Fast and syntax-aware semantic code pattern search for many languages: like grep but for code",
    cmdclass={"install": PostInstallCommand, "bdist_wheel": bdist_wheel},
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/returntocorp/semgrep",
    install_requires=[
        "colorama>=0.4.3",
        "pyyaml>=5.3",
        "requests>=2.22.0",
        "attrs>=19.3.0",
    ],
    entry_points={"console_scripts": ["semgrep=semgrep.__main__:main"]},
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    zip_safe=False,
)



import collections
import functools
import json
import multiprocessing
import re
import subprocess
import sys
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Dict
from typing import IO
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from semgrep.constants import PLEASE_FILE_ISSUE_TEXT
from semgrep.constants import SEMGREP_PATH
from semgrep.equivalences import Equivalence
from semgrep.evaluation import enumerate_patterns_in_boolean_expression
from semgrep.evaluation import evaluate
from semgrep.pattern import Pattern
from semgrep.pattern_match import PatternMatch
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.semgrep_types import BooleanRuleExpression
from semgrep.semgrep_types import OPERATORS
from semgrep.util import debug_print
from semgrep.util import INVALID_PATTERN_EXIT_CODE
from semgrep.util import partition
from semgrep.util import print_error
from semgrep.util import print_error_exit


def _offset_to_line_no(offset: int, buff: str) -> int:
    """
        Given string buffer return one indexed line number associated with byte offset
    """
    return buff.count("\n", 0, offset) + 1


def _offset_to_col_no(offset: int, buff: str) -> int:
    """
        Return one indexed col number associated with byte offset
    """
    return offset - buff.rfind("\n", 0, offset)


def get_re_matches(patterns_re: List[Tuple], path: Path) -> List[PatternMatch]:
    try:
        contents = path.read_text()
    except UnicodeDecodeError:
        debug_print(f"regex matcher skipping binary file at {path}")
        return []

    return [
        PatternMatch(
            {
                "check_id": pattern_id,
                "path": str(path),
                "start": {
                    "offset": match.start(),
                    "line": _offset_to_line_no(match.start(), contents),
                    "col": _offset_to_col_no(match.start(), contents),
                },
                "end": {
                    "offset": match.end(),
                    "line": _offset_to_line_no(match.end(), contents),
                    "col": _offset_to_col_no(match.end(), contents),
                },
                "extra": {"lines": [contents[match.start() : match.end()]]},
            }
        )
        for pattern_id, pattern in patterns_re
        for match in re.finditer(pattern, contents)
    ]


def get_target_files(
    targets: List[Path], exclude: List[str], include: List[str]
) -> List[Path]:
    if not include:
        # Default to all files
        include = ["*"]

    filepaths = [
        target
        for target in targets
        if target.is_file()
        and any(target.match(i) for i in include)
        and not any(target.match(e) for e in exclude)
    ]
    filepaths.extend(
        path
        for target in targets
        if target.is_dir()
        for path in target.rglob("*")
        if path.is_file()
        and any(path.match(i) for i in include)
        and not any(path.match(e) for e in exclude)
    )

    return filepaths


class CoreRunner:
    """
        Handles interactions between semgrep and semgrep-core

        This includes properly invoking semgrep-core and parsing the output
    """

    def __init__(
        self,
        allow_exec: bool,
        jobs: int,
        exclude: List[str],
        include: List[str],
        exclude_dir: List[str],
        include_dir: List[str],
    ):
        self._allow_exec = allow_exec
        self._jobs = jobs
        self._exclude = exclude
        self._include = include
        self._exclude_dir = exclude_dir
        self._include_dir = include_dir

    def _flatten_rule_patterns(self, rules: List[Rule]) -> Iterator[Pattern]:
        """
            Convert list of rules to format understandable by semgrep core
        """
        for rule_index, rule in enumerate(rules):
            flat_expressions = list(
                enumerate_patterns_in_boolean_expression(rule.expression)
            )
            for expr in flat_expressions:
                if not should_send_to_semgrep_core(expr):
                    continue

                yield Pattern(rule_index, expr, rule.severity, rule.languages)

    def _group_patterns_by_language(
        self, rules: List[Rule]
    ) -> Dict[str, List[Pattern]]:
        # a rule can have multiple patterns inside it. Flatten these so we can send semgrep a single yml file list of patterns
        patterns = list(self._flatten_rule_patterns(rules))
        by_lang: Dict[str, List[Pattern]] = collections.defaultdict(list)
        for pattern in patterns:
            for language in pattern.languages:
                by_lang[language].append(pattern)
        return by_lang

    def _semgrep_error_json_to_message_then_exit(
        self, error_json: Dict[str, Any]
    ) -> None:
        """
        See format_output_exception in semgrep O'Caml for details on schema
        """
        error_type = error_json["error"]
        if error_type == "invalid language":
            print_error_exit(f'invalid language {error_json["language"]}')
        elif error_type == "invalid pattern":
            print_error(
                f'invalid pattern "{error_json["pattern"]}": {error_json["message"]}'
            )
            sys.exit(INVALID_PATTERN_EXIT_CODE)
        # no special formatting ought to be required for the other types; the semgrep python should be performing
        # validation for them. So if any other type of error occurs, ask the user to file an issue
        else:
            print_error_exit(
                f'an internal error occured while invoking the semgrep engine: {error_type}: {error_json.get("message", "")}.\n\n{PLEASE_FILE_ISSUE_TEXT}'
            )

    def _flatten_all_equivalences(self, rules: List[Rule]) -> List[Equivalence]:
        """
        Convert all the equivalences defined in the rules into a single rule file
        """

        equivalences = []

        for rule in rules:
            try:
                equivalences.extend(rule.equivalences)
            except Exception as e:
                print_error_exit(f"could not get equivalences for rule {rule.id}; {e}")

        return equivalences

    def _write_equivalences_file(self, fp: IO, equivalences: List[Equivalence]) -> None:
        # I don't even know why this is a thing.
        # cf. https://stackoverflow.com/questions/51272814/python-yaml-dumping-pointer-references
        import yaml  # here for faster startup times

        yaml.SafeDumper.ignore_aliases = (  # type: ignore
            lambda *args: True
        )
        fp.write(yaml.safe_dump({"equivalences": [e.to_json() for e in equivalences]}))
        fp.flush()

    def _run_rules(
        self, rules: List[Rule], targets: List[Path]
    ) -> Tuple[Dict[Rule, Dict[Path, List[PatternMatch]]], List[Any]]:
        """
            Run all rules on targets and return list of all places that match patterns, ... todo errors
        """
        import yaml  # here for faster startup times

        outputs: List[PatternMatch] = []  # multiple invocations per language
        errors: List[Any] = []

        # This will need to be addressed in the future. Since we flatten all the patterns,
        # there's no way to tell semgrep which equivalences apply to which rules.
        # So, my approach here is a naive implementation which likewise flattens all equivalences...
        # Here there be dragons.... :-(
        #  .>   )\;`a__
        # (  _ _)/ /-." ~~
        #  `( )_ )/
        #   <_  <_ sb/dwb
        equivalences = self._flatten_all_equivalences(rules)
        with tempfile.NamedTemporaryFile("w") as equiv_fout:

            if equivalences:
                try:
                    self._write_equivalences_file(equiv_fout, equivalences)
                except Exception as e:
                    print_error(
                        f"could not write equivalences file. will continue without equivalences. {e}"
                    )
                    equivalences = []

            for language, all_patterns_for_language in self._group_patterns_by_language(
                rules
            ).items():
                # semgrep-core doesn't know about OPERATORS.REGEX - this is
                # strictly a semgrep Python feature. Regex filtering is
                # performed purely in Python code then compared against
                # semgrep-core's results for other patterns.
                patterns_regex, patterns = partition(
                    lambda p: p.expression.operator == OPERATORS.REGEX,
                    all_patterns_for_language,
                )
                if patterns_regex:
                    filepaths = get_target_files(targets, self._exclude, self._include)
                    patterns_json = [pattern.to_json() for pattern in patterns_regex]

                    try:
                        patterns_re = [
                            (pattern["id"], re.compile(pattern["pattern"]))
                            for pattern in patterns_json
                        ]
                    except re.error as err:
                        print_error_exit(f"invalid regular expression specified: {err}")

                    re_fn = functools.partial(get_re_matches, patterns_re)
                    with multiprocessing.Pool(self._jobs) as pool:
                        matches = pool.map(re_fn, filepaths)

                    outputs.extend(
                        single_match
                        for file_matches in matches
                        for single_match in file_matches
                    )

                patterns_json = [p.to_json() for p in patterns]
                # very important not to sort keys here
                yaml_as_str = yaml.safe_dump({"rules": patterns_json}, sort_keys=False)
                with tempfile.NamedTemporaryFile("w") as fout:
                    fout.write(yaml_as_str)
                    fout.flush()
                    cmd = [SEMGREP_PATH] + [
                        "-lang",
                        language,
                        f"-rules_file",
                        fout.name,
                    ]

                    if equivalences:
                        cmd += ["-equivalences", equiv_fout.name]
                    cmd += ["-j", str(self._jobs)]
                    cmd += [*self.targeting_options, *[str(path) for path in targets]]

                    try:
                        output = subprocess.check_output(cmd, shell=False)
                    except subprocess.CalledProcessError as ex:
                        try:
                            # see if semgrep output a JSON error that we can decode
                            semgrep_output = ex.output.decode("utf-8", "replace")
                            output_json = json.loads(semgrep_output)
                            if "error" in output_json:
                                self._semgrep_error_json_to_message_then_exit(
                                    output_json
                                )
                            else:
                                print_error(
                                    f"unexpected non-json output while invoking semgrep core with {' '.join(cmd)} \n {ex}"
                                )
                                print_error_exit(f"\n{PLEASE_FILE_ISSUE_TEXT}")
                                raise ex  # let our general exception handler take care of this
                        except Exception as e:
                            print_error(
                                f"non-zero return code while invoking semgrep with:\n\t{' '.join(cmd)}\n{ex} {e}"
                            )
                            print_error_exit(f"\n\n{PLEASE_FILE_ISSUE_TEXT}")
                    output_json = json.loads((output.decode("utf-8", "replace")))
                    errors.extend(output_json["errors"])
                    outputs.extend([PatternMatch(m) for m in output_json["matches"]])

        # group output; we want to see all of the same rule ids on the same file path
        by_rule_index: Dict[
            Rule, Dict[Path, List[PatternMatch]]
        ] = collections.defaultdict(lambda: collections.defaultdict(list))

        for pattern_match in outputs:
            rule_index = pattern_match.rule_index
            rule = rules[rule_index]
            by_rule_index[rule][pattern_match.path].append(pattern_match)

        return by_rule_index, errors

    def _resolve_output(
        self, outputs: Dict[Rule, Dict[Path, List[PatternMatch]]]
    ) -> Dict[Rule, List[RuleMatch]]:
        """
            Takes output of all running all patterns and rules and returns Findings
        """
        findings_by_rule: Dict[Rule, List[RuleMatch]] = {}

        for rule, paths in outputs.items():
            findings = []
            for filepath, pattern_matches in paths.items():
                if not rule.globs.match_path(filepath):
                    continue
                debug_print(f"-------- rule ({rule.id} ------ filepath: {filepath}")

                findings.extend(evaluate(rule, pattern_matches, self._allow_exec))

            findings_by_rule[rule] = dedup_output(findings)

        return findings_by_rule

    @property
    def targeting_options(self) -> Iterator[str]:
        """
            Yields include/exclude CLI options to call semgrep-core with.
            This is based on the arguments given to semgrep.
        """
        for pattern in self._exclude:
            yield from ["-exclude", pattern]
        for pattern in self._include:
            yield from ["-include", pattern]
        for pattern in self._exclude_dir:
            yield from ["-exclude-dir", pattern]
        for pattern in self._include_dir:
            yield from ["-include-dir", pattern]

    def invoke_semgrep(
        self, targets: List[Path], rules: List[Rule]
    ) -> Tuple[Dict[Rule, List[RuleMatch]], List[Any]]:
        """
            Takes in rules and targets and retuns object with findings
        """
        start = datetime.now()

        outputs, errors = self._run_rules(rules, targets)
        findings_by_rule = self._resolve_output(outputs)

        debug_print(f"semgrep ran in {datetime.now() - start}")

        return findings_by_rule, errors


def dedup_output(outputs: List[RuleMatch]) -> List[RuleMatch]:
    return list({uniq_id(r): r for r in outputs}.values())


def uniq_id(
    r: RuleMatch,
) -> Tuple[str, Path, Optional[int], Optional[int], Optional[int], Optional[int]]:
    start = r.start
    end = r.end
    return (
        r.id,
        r.path,
        start.get("line"),
        start.get("col"),
        end.get("line"),
        end.get("col"),
    )


def should_send_to_semgrep_core(expression: BooleanRuleExpression) -> bool:
    """
    don't send rules like "and-either" or "and-all" to semgrep
    """
    return (
        expression.pattern_id is not None
        and expression.operand is not None
        and (expression.operator != OPERATORS.WHERE_PYTHON)
    )



import subprocess
from pathlib import Path


def test_generate_config(run_semgrep_in_tmp):
    subprocess.check_output(["python3", "-m", "semgrep", "--generate-config"])
    run_semgrep_in_tmp(".semgrep.yml")



def foo(a, b):
    # ERROR: stupid $X == $X
    return a + b == a + b


def __eq__(a, b):
    # ERROR: stupid $X == $X (linter knows ok because method is named eq)
    x == x


# ERROR: stupid $X == $X (linter knows ok because method is assertTrue)
assertTrue(x == x)
